from typing import Sequence
from .mc_constraint import MCConstraint


class FixedIndices(MCConstraint):
    """Constrain a given set of indices during an MC run.
    Any suggested system changes by the MC algorithm are rejected if
    they invovle an index in the fixed indices.

    Parameters:

    fixed_indices: sequence of integers
        The indices of the atoms object which are to be fixed.
    """

    name = "FixedIndices"

    def __init__(self, fixed_indices: Sequence[int]):
        super().__init__()
        # We use a set for O(1) lookup time
        # Also, every entry is an index, and should therefore be an int.
        # Avoid floating point conversion issues with round
        self.fixed_basis = set(round(idx) for idx in fixed_indices)

    def __call__(self, system_changes):
        for idx, _, _ in system_changes:
            if idx in self.fixed_basis:
                return False
        return True



"""Monte Carlo method for ase."""
from typing import Tuple, Dict, Union
import sys
import datetime
import time
import logging
import random
from collections import Counter
import numpy as np
from ase import Atoms
from ase.units import kB
from clease.version import __version__
from clease.datastructures import SystemChange, SystemChanges
from .mc_evaluator import MCEvaluator
from .base import BaseMC
from .averager import Averager
from .bias_potential import BiasPotential
from .observers import MCObserver
from .trial_move_generator import TrialMoveGenerator, RandomSwap

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class Montecarlo(BaseMC):
    """Class for running Monte Carlo at a fixed composition (canonical).

    Args:
        system (Union[ase.Atoms, MCEvaluator]): Either an ASE Atoms object
            with an attached calculator, or a pre-initialized
            :class:`~clease.montecarlo.mc_evaluator.MCEvaluator`
            object.
        temp (float): Temperature of Monte Carlo simulation in Kelvin
        generator (TrialMoveGenerator, optional): A
            :class:`~clease.montecarlo.trial_move_generator.TrialMoveGenerator`
            object that produces trial moves. Defaults to None.
    """

    NAME = "MonteCarlo"

    def __init__(
        self,
        system: Union[Atoms, MCEvaluator],
        temp: float,
        generator: TrialMoveGenerator = None,
    ):

        super().__init__(system)
        self.T = temp

        if generator is None:
            self.generator = RandomSwap(self.atoms)
        else:
            self.generator = generator

        # List of observers that will be called every n-th step
        # similar to the ones used in the optimization routines
        self.observers = []
        self.bias_potentials = []

        self.current_step = 0
        self.num_accepted = 0
        self.status_every_sec = 30

        self.trial_move = []
        # We cannot cause an energy calculation trigger in init,
        # so we defer these quantities until needed.
        self.current_energy = None
        self.new_energy = None
        self.mean_energy = None
        self.energy_squared = None

        self.quit = False

    def update_current_energy(self):
        self.current_energy = self.evaluator.get_energy()
        logger.debug("Updating current energy to %s", self.current_energy)
        self.evaluator.reset()

    def _initialize_energies(self) -> None:
        """Initialize the energy quantities, causes a trigger of an energy evaluation."""

        self.update_current_energy()
        self.new_energy = self.current_energy

        # Create new averager objects, only the first time
        # (since they are initialized as None in __init__)
        if self.mean_energy is None:
            self.mean_energy = Averager(ref_value=self.current_energy)
        if self.energy_squared is None:
            self.energy_squared = Averager(ref_value=self.current_energy)

    def reset(self):
        """Reset all member variables to their original values."""
        logger.debug("Resetting.")
        for _, obs in self.observers:
            obs.reset()

        self.evaluator.reset()
        self._reset_internal_counters()

        if self.mean_energy is not None:
            self.mean_energy.clear()
        if self.energy_squared is not None:
            self.energy_squared.clear()

    def _reset_internal_counters(self):
        """Reset the step counters which are used internally"""
        self.current_step = 0
        self.num_accepted = 0

    def add_bias(self, potential: BiasPotential):
        """Add a new bias potential.

        Parameters:

        potential:
            Potential to be added
        """
        if not isinstance(potential, BiasPotential):
            raise TypeError("potential has to be of type BiasPotential")
        self.bias_potentials.append(potential)

    def attach(self, obs: MCObserver, interval: int = 1):
        """Attach observers to be called on a given MC step interval.

        Parameters:

        obs: MCObserver
            Observer to be added

        interval: int
            How often the observer should be called
        """
        if not obs.interval_ok(interval):
            name = type(obs).__name__
            raise ValueError(f"Invalid interval for {name}. Check docstring of the observer.")

        self.observers.append((interval, obs))

    def initialize_run(self):
        """Prepare MC object for a new run."""
        logger.debug("Initializing run")
        self.generator.initialize(self.atoms)

        # Ensure the evaluator is properly synchronized.
        self.evaluator.synchronize()

        # Initialize/update relevant energy quantities
        self._initialize_energies()

        # Reset the internal step counters
        self._reset_internal_counters()

    def run(self, steps: int = 100):
        """Run Monte Carlo simulation.

        Parameters:

        steps: int
            Number of steps in the MC simulation
        """

        logger.info("Starting MC run with %d steps.", steps)
        # Check the number of different elements are correct to avoid
        # infinite loops
        self.initialize_run()

        start = time.perf_counter()
        prev = self.current_step
        info_enabled = logger.isEnabledFor(logging.INFO)  # Do we emit INFO logs?
        while self.current_step < steps:
            E, _ = self._mc_step()

            self.mean_energy += E
            self.energy_squared += E**2

            # We only want to do this calculation if logging is enabled for INFO.
            if info_enabled and time.perf_counter() - start > self.status_every_sec:
                ms_per_step = 1000.0 * self.status_every_sec / (self.current_step - prev)
                accept_rate = self.num_accepted / self.current_step
                logger.info(
                    "%d of %d steps. %.2f ms per step. Acceptance rate: %.2f",
                    self.current_step,
                    steps,
                    ms_per_step,
                    accept_rate,
                )
                prev = self.current_step
                start = time.perf_counter()

            if self.quit:
                logger.warning("Quit signal detected. Breaking the MC loop.")
                break

        logger.info("Reached maximum number of steps (%d mc steps)", steps)

    @property
    def meta_info(self):
        """Return dict with meta info."""
        ts = time.time()
        st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H:%M:%S")
        clease_version = str(__version__)
        v_info = sys.version_info
        meta_info = {
            "timestamp": st,
            "python_version": f"{v_info.major}.{v_info.minor}.{v_info.micro}",
            "clease_version": clease_version,
        }
        return meta_info

    def get_thermodynamic_quantities(self):
        """Compute thermodynamic quantities."""
        quantities = {}
        mean_energy = self.mean_energy.mean
        quantities["energy"] = mean_energy
        mean_sq = self.energy_squared.mean
        quantities["heat_capacity"] = (mean_sq - mean_energy**2) / (kB * self.T**2)
        quantities["energy_var"] = mean_sq - mean_energy**2
        quantities["temperature"] = self.T
        at_count = self.count_atoms()
        for key, value in at_count.items():
            name = f"{key}_conc"
            conc = float(value) / len(self.atoms)
            quantities[name] = conc

        # Add some more info that can be useful
        quantities.update(self.meta_info)

        # Add information from observers
        for obs in self.observers:
            quantities.update(obs[1].get_averages())
        return quantities

    def _calculate_step(self, system_changes: SystemChanges):
        """Calculate energies given a step, and decide if we accept the step.

        Returns boolean if system changes are accepted.

        Parameters:

        system_changes: list
            List with system changes
        """
        logger.debug("Applying changes to the system.")
        self.evaluator.apply_system_changes(system_changes)
        # Evaluate the energy quantity after applying the changes to the system.
        self.new_energy = self.evaluator.get_energy(applied_changes=system_changes)

        # NOTE: As this is called after calculate, the changes has
        # already been introduced to the system
        for bias in self.bias_potentials:
            self.new_energy += bias(system_changes)
        logger.debug(
            "Current energy: %.3f eV, new energy: %.3f eV",
            self.current_energy,
            self.new_energy,
        )
        accept = self._do_accept(self.current_energy, self.new_energy)
        logger.debug("Change accepted? %s", accept)

        if accept:
            # Changes accepted, finalize evaluator.
            self.evaluator.keep_system_changes(system_changes)
        else:
            # Undo changes
            logger.debug("Change rejected, undoing system changes.")
            self.evaluator.undo_system_changes(system_changes)

        return accept

    def _do_accept(self, current_energy: float, new_energy: float) -> bool:
        """Decide if we accept a state, based on the energies.

        Return a bool on whether the move was accepted.

        Parameters:

        :param current_energy: Energy of the current configuration
        :param new_energy: Energy of the new configuration.
        """
        # Standard Metropolis acceptance criteria
        if new_energy < current_energy:
            return True
        kT = self.T * kB
        energy_diff = new_energy - current_energy
        probability = np.exp(-energy_diff / kT)
        logger.debug(
            "Energy difference: %.3e. Calculated probability: %.3f",
            energy_diff,
            probability,
        )
        return random.random() <= probability

    def _move_accepted(self, system_changes: SystemChanges) -> SystemChanges:
        logger.debug("Move accepted, updating things")
        self.num_accepted += 1
        self.generator.on_move_accepted(system_changes)
        self.current_energy = self.new_energy
        return system_changes

    def _move_rejected(self, system_changes: SystemChanges) -> SystemChanges:
        logger.debug("Move rejected, undoing system changes: %s", system_changes)
        self.generator.on_move_rejected(system_changes)

        # Move rejected, no changes are made
        system_changes = [
            SystemChange(
                index=change.index,
                old_symb=change.old_symb,
                new_symb=change.old_symb,
                name=change.name,
            )
            for change in system_changes
        ]
        logger.debug("Reversed system changes: %s", system_changes)
        return system_changes

    def count_atoms(self) -> Dict[str, int]:
        """Count the number of each element."""
        return dict(Counter(self.atoms.symbols))

    def _mc_step(self) -> Tuple[float, bool]:
        """Make one Monte Carlo step by swithing two atoms."""
        self.current_step += 1

        system_changes = self.generator.get_trial_move()
        self.trial_move = system_changes

        # Calculate step, and whether we accept it
        move_accepted = self._calculate_step(system_changes)

        updater = self._move_accepted if move_accepted else self._move_rejected
        system_changes = updater(system_changes)

        # Execute all observers
        self.execute_observers(system_changes)
        return self.current_energy, move_accepted

    def execute_observers(self, system_changes: SystemChanges):
        for interval, obs in self.observers:
            if self.current_step % interval == 0:
                logger.debug(
                    "Executing observer %s at step %d with interval %d.",
                    obs.name,
                    self.current_step,
                    interval,
                )
                obs(system_changes)



import logging
from collections import defaultdict
from typing import List, Dict, Set, Sequence, Union
from datetime import datetime
from packaging.version import parse, Version
import numpy as np
import ase
from ase.db import connect
from .version import __version__  # Avoid triggering other imports

logger = logging.getLogger(__name__)

# Global switch for emergency disabling checking of version checks
REQUIRE_COMPATIBLE_TABLE_VERISON = True


def toggle_require_compatible_table_version() -> None:
    """Toggle to disable/enable the "require_reconfigure_table" function."""
    global REQUIRE_COMPATIBLE_TABLE_VERISON  # pylint: disable=global-statement
    new = not REQUIRE_COMPATIBLE_TABLE_VERISON
    logger.info(
        'Toggling global "REQUIRE_COMPATIBLE_TABLE_VERISON" from %s to %s.',
        REQUIRE_COMPATIBLE_TABLE_VERISON,
        new,
    )
    REQUIRE_COMPATIBLE_TABLE_VERISON = new


class MetaTableKeys:
    """Container for keys in the meta table"""

    # Key where we store what version a database has been configured on
    CLEASE_CONFIGURE_VERSION = "clease_config_version"
    # External table with clease table metadata
    CLEASE_META_TABLE = "clease_metadata"
    TIME = "time"
    ENCODE_DELIMITER = "__"  # Delimiter splitting encoding a table-name/table-key pair


class OutOfDateTable(Exception):
    """Exception for having an out-of-date table which requires an update/reconfigure"""


class InvalidMetaKey(Exception):
    """Exception when passing an invalid key in the metadata"""


# Type aliasing
_TABLE = Dict[str, Union[int, float, str, np.integer, np.floating]]  # Alias for a table
_ROW = ase.db.row.AtomsRow
_CONNECTION = ase.db.core.Database


def get_metadata(ase_connection: _CONNECTION, *ids) -> _TABLE:
    """Get a the metadata from a a database, without using the ASE API,
    as that causes loading of the entire database, which can be slow
    for larger databases. Therefore, we only extract what we need.
    If no metadata is present in the database, an empty dictionary is returned.

    Returns the decoded version of the meta data, i.e. a nested dictionary
    with metadata for each table.

    :param ase_connection: The ASE connection to the database.
    :param *ids: One or more ID's to get meta data for. Must be integers.
    """
    table_name = MetaTableKeys.CLEASE_META_TABLE
    # pylint: disable=protected-access
    if not ase_connection._external_table_exists(table_name):
        logger.debug("No metadata table was found")
        return {}
    sql = f"SELECT key,value,id FROM {table_name}"
    logger.info("Grabbing metadata with querry %s", sql)
    with ase_connection.managed_connection() as con:
        cur = con.cursor()
        cur.execute(sql)
        metadata = defaultdict(dict)
        for key, value, row_id in cur.fetchall():
            if row_id in ids:
                metadata[row_id].update({key: value})
    logger.debug("%s", metadata)
    return {k: decode_meta(v) for k, v in metadata.items()}


def require_reconfigure_table(ase_connection: _CONNECTION, table_name: str, *ids) -> bool:
    """Core function which checks the version the database was configured with.
    Returns a boolean indicating if the correlation functions of the
    database needs to be reconfigured, based on verison changes.

    The "REQUIRE_COMPATIBLE_TABLE_VERISON" global can be set to ``False`` to disable
    this functionality.

    :param ase_connection: The ASE connection to the database.
    :param table_name: ``str``, name of the table to check if
        it requires reconfiguration.
    :param *ids: One or more ID's to get meta data for. Must be integers.
    """

    if REQUIRE_COMPATIBLE_TABLE_VERISON is False:
        # Short circuit, for debugging/emergency situations
        return False

    if len(ids) == 0:
        raise ValueError("Must pass at least 1 id to test for reconfigure")

    metadata_many = get_metadata(ase_connection, *ids)
    for uid in ids:
        meta = metadata_many.get(uid, {})

        table = meta.get(table_name, {})

        value = table.get(MetaTableKeys.CLEASE_CONFIGURE_VERSION, None)
        if value is None:
            # Configured prior to 0.10.1
            # or not a CLEASE database
            return True
        vers = parse(value)
        if vers < Version("0.10.1"):
            # 0.10.1 introduced a change, requiring a reconfigure
            return True
    return False  # We're up-to-date


def _make_meta_for_table(**kwargs) -> _TABLE:
    """Update the metadata for a table. Preserves any other metadata placed into
    the table. If no table is given, a new one is created."""

    meta = {}
    now = datetime.now().strftime("%H:%M:%S %d-%m-%Y")  # Store a timestamp of the config
    version = str(__version__)
    meta.update(
        {
            MetaTableKeys.TIME: now,
            MetaTableKeys.CLEASE_CONFIGURE_VERSION: version,
            **kwargs,
        }
    )
    return meta


def _check_key(key, delim=MetaTableKeys.ENCODE_DELIMITER) -> None:
    """Ensure that a key does not contain the delimiter used for encoding"""
    if delim in key:
        raise InvalidMetaKey(f'Encoding delimiter "{delim}" cannot be present in key, got: {key}')


def encode_meta_key(table_name, key, delim=MetaTableKeys.ENCODE_DELIMITER):
    """Encode a key for metadata, combining a table name with the key.

    E.g.
    >>> from clease.db_util import encode_meta_key
    >>> encode_meta_key('poly_cf', 'time')
    'poly_cf__time'

    :param table_name: String, name of the table
    :param key: String, the key to encode
    :param delim: Optional, what delimiter to use. Should usually not be changed.
        Defaults to '__'.
    """
    # Check the delimiter is not present in the keys
    for k in (table_name, key):
        _check_key(k, delim=delim)

    return delim.join((table_name, key))


def decode_meta_key(key, delim=MetaTableKeys.ENCODE_DELIMITER):
    """Decode a key which has been encoded using func:`encode_meta_key`.

    E.g.
    >>> from clease.db_util import decode_meta_key
    >>> decode_meta_key('poly_cf__time')
    ('poly_cf', 'time')

    :param key: String, key to decode.
    :param delim: Optional, what delimiter to use. Should usually not be changed.
        Defaults to '__'.
    """
    if delim not in key:
        raise InvalidMetaKey(f'Key "{key}" does not contain delimiter {delim}')
    table_name, table_key = key.split(delim)
    return table_name, table_key


def decode_meta(meta: dict) -> dict:
    """Decode an entire meta-data dictionary encoded using :func:`encode_meta`.

    The dictionary is returned in a nested format.
    """
    decoded = defaultdict(dict)
    for key, value in meta.items():
        table_name, table_key = decode_meta_key(key)
        decoded[table_name].update({table_key: value})
    return dict(decoded)


def encode_meta(meta: dict) -> dict:
    """Encodes a nested dictionary, such that it becomes a flat dictionary."""
    encoded = {}
    for table_name, table in meta.items():
        for key, value in table.items():
            encoded_key = encode_meta_key(table_name, key)
            encoded[encoded_key] = value
    return encoded


def make_meta_table(table_name: str, **kwargs) -> _TABLE:
    """Create a table with CLEASE meta data, such as the version the clease version
    the table was created with.

    Preserves any other metadata in the table, besides :func:`_make_meta_for_table`
    will update.
    """

    return {table_name: _make_meta_for_table(**kwargs)}


def new_row_with_many_tables(
    ase_connection: _CONNECTION,
    atoms,
    table_names: Sequence[str],
    tables: Sequence[_TABLE],
    **kwargs,
) -> int:
    """Write a new row to a database with multiple external tables. Adds metadata for each table.

    Returns the ID of the new row in the database.

    :param ase_connection: An open ASE database connection
    :param atoms: The atoms object to write the row for
    :param table_names: Names of the tables to be added
    :param tables: Sequence of the tables. Order must be the same as ``table_names``
    :param kwargs: Any other key-value pair to be added to the row.
        See documentation for the `ASE database <https://wiki.fysik.dtu.dk/ase/ase/db/db.html>`_.
    """
    if not len(table_names) == len(tables):
        raise ValueError(
            (
                f"Number of table names should match number of tables. Got {len(table_names)} "
                f"table names, but {len(tables)} tables."
            )
        )

    metatable = {}
    for table_name in table_names:
        metatable.update(make_meta_table(table_name))

    tables = dict(zip(table_names, tables))
    tables.update({MetaTableKeys.CLEASE_META_TABLE: encode_meta(metatable)})
    uid = ase_connection.write(atoms, external_tables=tables, **kwargs)
    return uid


def new_row_with_single_table(
    ase_connection: _CONNECTION,
    atoms: ase.Atoms,
    table_name: str,
    table: _TABLE,
    **kwargs,
) -> int:
    """Write a new row to a database with a single external table. Adds metadata to the table.

    See :func:`new_row_with_many_tables` for more details.
    """
    return new_row_with_many_tables(ase_connection, atoms, [table_name], [table], **kwargs)


def update_table(ase_connection: _CONNECTION, row_id: int, table_name: str, table: _TABLE) -> None:
    """Configure a row in an ASE database with an external table,
    and add a corresponding clease metadata table."""
    # Get the current row from the database connection,
    # as we need to know if it already contains metadata
    metatable = make_meta_table(table_name)
    meta = {MetaTableKeys.CLEASE_META_TABLE: encode_meta(metatable)}
    tables = dict(**{table_name: table}, **meta)
    logger.debug(
        'Updating table "%s" for row id %d with keys: %s',
        table_name,
        row_id,
        tables.keys(),
    )
    ase_connection.update(row_id, external_tables=tables)


def get_all_cf_names(db_name: str, tab_name: str) -> Set[str]:
    """
    Return a list with all correlation function names

    Parameters:

    db_name: Name of the database

    tab_name: Name of the table in the database where correlation functions
        are stores
    """
    with connect(db_name) as db:
        query = f"SELECT key FROM {tab_name}"
        cur = db.connection.cursor()
        db_names = set()
        cur.execute(query)
        for row in cur.fetchall():
            db_names.add(row[0])
    return db_names


def get_all_cf(db_name: str, tab_name: str, db_id: int) -> Dict[str, float]:
    """
    Return all correlation functions associated with an entry in the database

    Parameters:

    db_name: Name of the database

    tab_name: Name of the table where the correlation functions are stored

    db_id: ID of the entry
    """
    with connect(db_name) as db:
        query = f"SELECT key, value FROM {tab_name} WHERE id=?"
        cur = db.connection.cursor()
        cur.execute(query, (db_id,))
        cf = {}
        for row in cur.fetchall():
            cf_name, cf_value = row
            cf[cf_name] = cf_value
    return cf


def get_cf_tables(db_name: str) -> List[str]:
    """
    Return a list with table names that contain correlation functions

    Parameter:

    db_name: Name of the database
    """
    db = connect(db_name)
    ext_tab = db._get_external_table_names()  # pylint: disable=protected-access
    cf_tables = [n for n in ext_tab if n.endswith("_cf")]
    return cf_tables



from typing import NamedTuple, Sequence
import typing
import ase
from ase.data import atomic_numbers

__all__ = ("SystemChange", "SystemChanges")


def _apply_symbol_to_index(atoms: ase.Atoms, index: int, symbol: str) -> None:
    """Faster method for changing a single symbol at a given index, since it short-circuits
    the whole atoms.symbols mechanism, since we already know we only want to change 1 symbol.

    Roughly equivalent to

    atoms.symbols[index] = new_sybmol
    """
    number = atomic_numbers[symbol]
    atoms.numbers[index] = number


class SystemChange(NamedTuple):
    index: int
    old_symb: str
    new_symb: str
    name: str

    def apply_change(self, atoms: ase.Atoms) -> None:
        """Apply the change to the atoms object. The atoms object is mutated in place.
        Note, no check is performed to ensure that the 'old_symb' is correct.

        :param atoms: ASE Atoms object to be mutated.
        """
        _apply_symbol_to_index(atoms, self.index, self.new_symb)

    def undo_change(self, atoms: ase.Atoms) -> None:
        """Undo the change to the atoms object. The atoms object is mutated in place.
        Note, that no check is performed to ensure that the changed symbol is what is expected.

        :param atoms: ASE Atoms object to be mutated.
        """
        _apply_symbol_to_index(atoms, self.index, self.old_symb)


# Type alias for multiple SystemChange objects
SystemChanges = Sequence[SystemChange]


def get_inverted_changes(
    system_changes: SystemChanges,
) -> typing.Iterator[SystemChange]:
    """Invert system changes by doing old_symbs -> new_symbs."""
    yield from (
        SystemChange(
            index=change.index,
            old_symb=change.new_symb,
            new_symb=change.old_symb,
            name=change.name,
        )
        for change in system_changes
    )



from .clease import Clease, MovedIgnoredAtomError
from .clease_vol_dep import CleaseVolDep
from .util import attach_calculator, get_ce_energy

__all__ = [
    "Clease",
    "MovedIgnoredAtomError",
    "attach_calculator",
    "get_ce_energy",
    "CleaseVolDep",
]



from typing import Dict, Optional, Iterable, Set, List, Union, Sequence
from ase import Atoms
from ase.calculators.calculator import Calculator
from clease.datastructures import SystemChange
from clease.settings import ClusterExpansionSettings
from .clease import Clease


class CleaseVolDep(Clease):
    """
    Calculator that can be used together with volume dependent ECIs

    :param settings: ClusterExpansionSettings object used to construct the
        cluster expansion
    :param eci: Dictionary with ECI. The keys in the dictionary should have the
        following form: c<size>_<diameter>_<id>_<decoration_number>_V<power>.
        Example: if the contribution to the total energy from the
        nearest neighbour pair cluster has the mathematical form

        (A + B*V + C*V^2)*CF,

        where CF is the nearest neighbour pair correlation function, the
        following coefficients should be in the ECI dict (the id is 0 and
        decoration number 00 in this example. They could be different for other
        correlation functions)

        {
            'c2_d0000_0_00_V0': A,
            'c2_d0000_0_00_V1': B,
            'c2_d0000_0_00_V2': C
        }

        The total energy is given by a sum of terms of the form depicted above.
        Example: If the total energy depends on the one-body, nearest neighbour
        pair cluster and shorted three body cluster with correlation functions
        CF1, CF2 and CF3, respectively. The expression for the total energy is

        E = (A1 + B1*V + C1*V^2)*CF1 + (A2 + B2*V + C2*V^2)*CF2 +
            (A3 + B3*V + C3*V^2)*CF3

        the ECI dictionary required to represent this expression is

        {
            'c1_0_V0': A1,
            'c1_0_V1': B1,
            'c1_0_V3': C1,
            'c2_d0000_0_00_V0': A2,
            'c2_d0000_0_00_V1': B2,
            'c2_d0000_0_00_V2': C2,
            'c3_d0000_0_000_V0': A3,
            'c3_d0000_0_000_V1': B3,
            'c3_d0000_0_000_V2': C3,
        }

        again the id has been set to 0 and the decoration numbers are only 00
        (and 000), which will vary for different types of correlation functions.
    :param vol_coeff: Expansion coefficients used to cluster expand the volume.
        The volume coefficients should not have the V<power> tag at the end.
        Example: If the volume can be described by the following mathematical
        form

        V = A + B*CF1 + C*CF2,

        where CF1 is the correlation function of one-body clusters and CF2 is
        the correlation function of the nearest neighbour pair interacation,
        the passed dictionary should be

        {
            'c0': A,
            'c1_0': B,
            'c2_d0000_0_00': C
        }
    :param init_cf: Correlation functions of the initial atoms object. They can
        for instance be obtained via `clease.CorrFunction`. If not passed,
        they are calculated from scratch when an atoms object is attached.
    """

    def __init__(
        self,
        settings: ClusterExpansionSettings,
        eci: Dict[str, float],
        vol_coeff: Dict[str, float],
        init_cf: Optional[Dict[str, float]] = None,
    ):

        if not eci_format_ok(eci.keys()):
            raise ValueError(f"Invalid format of ECI names. Got\n{eci.keys()}")

        if not vol_coeff_format_ok(vol_coeff.keys()):
            raise ValueError(
                "Invalid format of volume coefficient names. " f"Got\n{vol_coeff.keys()}"
            )

        cf_to_track = unique_eci_names_no_vol(eci.keys())
        cf_to_track = cf_to_track.union(set(vol_coeff.keys()))

        # Create a set of empty ECI and pass to the parent class.
        # This makes sure that all nessecary correlation functions
        # are tracked.
        eci_track = {k: 0.0 for k in cf_to_track}

        # Transfer the volume independent part to the parent calculator
        for k, v in eci.items():
            if k.endswith("V0"):
                key = k.rpartition("_V0")[0]
                eci_track[key] = v

        Clease.__init__(self, settings, eci_track, init_cf=init_cf)
        self.eci_with_vol = eci
        self.vol_coeff = vol_coeff
        self.max_power = max(int(k[-1]) for k in self.eci_with_vol.keys())

    def get_volume(self, cf: Optional[Dict[str, float]] = None) -> float:
        """
        Returns the volume per atom

        :param cf: Correlation functions. If not given, the correlation
            functions are updated to match the current state of the attached
            atoms object
        """
        if cf is None:
            self.update_cf(None)
            cf = self.get_cf()
        vol = sum(self.vol_coeff[k] * cf[k] for k in self.vol_coeff)
        self.results["volume_per_atom"] = vol
        self.results["volume"] = vol * len(self.atoms)
        return vol

    def get_pressure(self, cf: Optional[Dict[str, float]] = None) -> float:
        """
        Return the pressure

        :param cf: Correlation functions. If not given, the correlation
            functions are updated to match the current state of the attached
            atoms object
        """
        if cf is None:
            self.update_cf(None)
            cf = self.get_cf()

        vol = self.get_volume(cf)
        P = sum(
            p * self.eci_with_vol.get(k + f"_V{p}", 0.0) * cf[k] * vol ** (p - 1)
            for k in cf.keys()
            for p in range(1, self.max_power + 1)
        )
        self.results["pressure"] = P
        return P

    def get_bulk_modulus(self, cf: Optional[Dict[str, float]] = None) -> float:
        """
        Return the bulk modulus of the current atoms object

        :param cf: Correlation functions. If not given, the correlation
            functions are updated to match the current state of the attached
            atoms object
        """
        if cf is None:
            self.update_cf(None)
            cf = self.get_cf()

        vol = self.get_volume(cf)
        B = vol * self._d2EdV2(cf, vol)
        self.results["bulk_mod"] = B
        return B

    def _d2EdV2(self, cf: Dict[str, float], vol: float) -> float:
        """
        Return the double derivative of the energy with respect
        to the volume

        :param cf: Correlation functions
        :param vol: Volume
        """
        return sum(
            p * (p - 1) * self.eci_with_vol.get(k + f"_V{p}", 0.0) * cf[k] * vol ** (p - 2)
            for k in cf.keys()
            for p in range(2, self.max_power + 1)
        )

    def _d3EdV3(self, cf: Dict[str, float], vol: float) -> float:
        """
        Return the third derivative of the energy with respect to
        the volume

        :param cf: Correlation functions
        :param vol: Volume
        """
        return sum(
            p
            * (p - 1)
            * (p - 2)
            * self.eci_with_vol.get(k + f"_V{p}", 0.0)
            * cf[k]
            * vol ** (p - 3)
            for k in cf.keys()
            for p in range(3, self.max_power + 1)
        )

    def get_dBdP(self, cf: Dict[str, float] = None) -> float:
        """
        Return the pressure derivative of the bulk modulus of the
        current structure.

        :param cf: Correlation functions. If not given, the correlation
            functions are updated to match the current state of the attached
            atoms object
        """
        if cf is None:
            self.update_cf(None)
            cf = self.get_cf()
        vol = self.get_volume(cf)
        return -1.0 - vol * self._d3EdV3(cf, vol) / self._d2EdV2(cf, vol)

    def get_energy_given_change(self, system_changes: Sequence[SystemChange], keep_changes=False):
        """
        Update correlation functions given the change
        """
        with self.with_system_changes(system_changes) as keeper:
            keeper.keep_changes = keep_changes
            cf = self.get_cf()
            vol = self.get_volume(cf)
            energy = sum(
                self.eci_with_vol.get(k + f"_V{p}", 0.0) * cf[k] * vol**p
                for k in cf.keys()
                for p in range(self.max_power + 1)
            )
            self.energy = energy * len(self.atoms)
            self.results["energy"] = self.energy
            return self.energy

    def calculate(
        self,
        atoms: Atoms,
        properties: List[str],
        system_changes: Union[Sequence[SystemChange], None],
    ) -> float:
        """Calculate the energy of the passed Atoms object.

        If accept=True, the most recently used atoms object is used as a
        reference structure to calculate the energy of the passed atoms.
        Returns energy.

        Parameters:

        :param atoms: ASE Atoms object

        :param system_changes:
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        Calculator.calculate(self, atoms)
        self.update_cf()
        return self.get_energy_given_change([], keep_changes=True)

    def _on_eci_changed(self):
        """
        Callback triggered by the parent class when the ECIs are changed.
        """
        for k, v in self.eci.items():
            vol_dep_key = k + "_V0"
            self.eci_with_vol[vol_dep_key] = v


def unique_eci_names_no_vol(eci_names: Iterable[str]) -> Set[str]:
    """
    Return a set with the unique ECI names without any volume tag
    """
    eci_no_vol = set(k.rpartition("_V")[0] for k in eci_names)
    return eci_no_vol


def eci_format_ok(eci_names: Iterable[str]) -> bool:
    """
    Check that all ECIs are formatted correctly.
    """
    for name in eci_names:
        valid = name.startswith("c") and "_V" in name
        if not valid:
            return False
    return True


def vol_coeff_format_ok(vol_coeff_names: Iterable[str]) -> bool:
    """
    Check that the volume coefficients are valid
    """
    for name in vol_coeff_names:
        valid = name.startswith("c") and "_V" not in name
        if not valid:
            return False
    return True



import logging
from typing import Tuple, List, Dict, Set, Optional, Sequence
import sqlite3
from abc import ABC, abstractmethod
from itertools import combinations_with_replacement as cwr
import numpy as np
from ase.db import connect
from clease import db_util
from clease.tools import add_file_extension, sort_cf_names, get_ids

logger = logging.getLogger(__name__)

__all__ = (
    "InconsistentDataError",
    "DataManager",
    "CorrFuncEnergyDataManager",
    "CorrelationFunctionGetter",
    "CorrFuncVolumeDataManager",
    "CorrelationFunctionGetterVolDepECI",
    "FinalStructPropertyGetter",
    "make_corr_func_data_manager",
)


class InconsistentDataError(Exception):
    """Data is inconsistent"""


class PropertyGetter(ABC):
    """Base class for getting a property from the database"""

    def __init__(self, db_name: str):
        self.db_name = db_name

    @abstractmethod
    def get_property(self, ids: Sequence[int]) -> np.ndarray:
        """Return the properties of all structures which matches the criteria"""

    def connect(self, **kwargs):
        """Return an ASE connection to the database"""
        return connect(self.db_name, **kwargs)


class FeatureGetter(PropertyGetter, ABC):
    """A base class for getting features, e.g. correlation functions"""

    @property
    @abstractmethod
    def names(self) -> Sequence[str]:
        """Return a name for each feature"""


class TargetGetter(PropertyGetter, ABC):
    """A base class for getting targets, e.g. energy per atom"""

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the name of the target property"""


class DataManager(ABC):
    """
    DataManager is a class for extracting data from CLEASE databases to be
    used to fit ECIs

    :param db_name: Name of the database
    """

    def __init__(self, db_name: str):
        self.db_name = db_name
        self._X = None
        self._y = None
        self._feat_names = None
        self._target_name = None

    @abstractmethod
    def get_data(self, select_cond: List[tuple]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return the design matrix X and the target data y
        """

    def _get_data_from_getters(
        self, select_cond: List[tuple], feature_getter: FeatureGetter, target_getter: TargetGetter
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return the design matrix X and the target data y

        :param select_cond: List of tuples with selection conditions
            (e.g. [('converged', '='' True)])

        :param feature_getter:
            Instance of a FeatureGetter object that returns design matrix.

        :param target_getter:
            Instance of a TargetGetter object that extracts the target value (e.g. energy per atom)

        Example:

        >>> import clease
        >>> from clease.data_manager import (
        ... CorrelationFunctionGetter, FinalStructEnergyGetter,
        ... DataManager)
        >>> from ase.db import connect
        >>> from ase import Atoms
        >>> from ase.calculators.emt import EMT
        >>> db_name = 'somedb.db'
        >>> db = connect(db_name)  # Make sure that the DB exists
        >>> clease.db_util.new_row_with_single_table(db, Atoms('Cu'),
        ...  'polynomial_cf', {'c0': 1.0}, final_struct_id=2, converged=True)
        1
        >>> final = Atoms('Cu')
        >>> final.calc = EMT()
        >>> _ = final.get_potential_energy()
        >>> db.write(final)
        2
        >>> feat_getter = CorrelationFunctionGetter(db_name, 'polynomial_cf',
        ... ['c0', 'c1_0', 'c2_d0000_0_00'], )
        >>> targ_getter = FinalStructEnergyGetter(db_name)
        >>> manager = DataManager(db_name)
        >>> X, y = manager.get_data([('converged', '=', 1)], feat_getter,
        ... targ_getter)
        >>> import os
        >>> os.remove(db_name)
        """
        ids = get_ids(select_cond, self.db_name)

        # Extract design matrix and the target values
        cfm = feature_getter.get_property(ids)
        target = target_getter.get_property(ids)

        self._X = np.array(cfm, dtype=float)
        self._y = np.array(target)
        self._feat_names = feature_getter.names
        self._target_name = target_getter.name

        nrows = self._X.shape[0]
        num_data = len(self._y)
        if nrows != num_data:
            raise InconsistentDataError(
                f"Num. rows in deisgn matrix {nrows}."
                f"Num. data points {num_data}\n"
                "This is normally caused by a non-converged structure "
                "being extracted from database. Check that all structures"
                "extracted from DB using your query has a corresponding"
                "final structure."
            )
        return self._X, self._y

    def to_csv(self, fname: str):
        """
        Export the dataset used to fit a model y = Xc where y is typically the
        DFT energy per atom and c is the unknown ECIs. This function exports
        the data to a csv file with the following format

        # ECIname_1, ECIname_2, ..., ECIname_n, E_DFT
        0.1, 0.4, ..., -0.6, -2.0
        0.3, 0.2, ..., -0.9, -2.3

        thus each row in the file contains the correlation function values and
        the corresponding DFT energy value.

        :param fname: Filename to write to. Typically this should end with .csv
        """
        if self._X is None:
            return
        fname = add_file_extension(fname, ".csv")
        header = ",".join(self._feat_names) + f",{self._target_name}"
        data = np.hstack((self._X, np.reshape(self._y, (len(self._y), -1))))
        np.savetxt(fname, data, delimiter=",", header=header)
        logger.info("Dataset exported to %s", fname)

    def get_matching_names(self, pattern: str) -> List[str]:
        """
        Get names that matches pattern

        :param pattern: Pattern which the string should contain.

        Example:

        If the names are ['abc', 'def', 'gbcr'] and the passed pattern
        is 'bc', then ['abc', 'gbcr'] will be returned
        """
        return [n for n in self._feat_names if pattern in n]

    def get_cols(self, names: List[str]) -> np.ndarray:
        """
        Get all columns corresponding to the names

        :pram names: List of names (e.g. ['c0', 'c1_1'])
        """
        name_idx = {n: i for i, n in enumerate(self._feat_names)}
        indices = [name_idx[n] for n in names]
        return self._X[:, indices]

    def groups(self) -> List[int]:
        """
        Returns the group of each item in the X matrix. In the top-level
        DataManager it is assumed that each row in the X matrix constitutes
        its own group. But this method may be overrided in child classes.
        """
        if self._X is None:
            return []
        return list(range(self._X.shape[0]))


class CorrFuncEnergyDataManager(DataManager):
    """
    CorrFuncFinalEnergyDataManager is a convenience class provided
    to handle the standard case where the features are correlation functions
    and the target is the DFT energy per atom

    :param db_name: Name of the database being passed

    :param cf_names: List with the correlation function names to extract

    :param tab_name: Name of the table where the correlation functions are
        stored

    :param order: Order of the correlation function. Default 1.
    """

    def __init__(
        self,
        db_name: str,
        tab_name: str,
        cf_names: Optional[List[str]] = None,
        order: int = 1,
    ) -> None:
        super().__init__(db_name)
        self.tab_name = tab_name
        self.cf_names = cf_names
        self.order = order

    def get_data(self, select_cond: List[tuple]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return X and y, where X is the design matrix containing correlation
        functions and y is the DFT energy per atom.

        :param select_cond: List with select conditions for the database
            (e.g. [('converged', '=', True)])
        """
        # pylint: disable=arguments-differ
        return self._get_data_from_getters(
            select_cond,
            CorrelationFunctionGetter(self.db_name, self.tab_name, self.cf_names, order=self.order),
            FinalStructEnergyGetter(self.db_name),
        )


class CorrFuncPropertyDataManager(DataManager):
    """
    CorrFuncFinalEnergyDataManager is a convenience class provided
    to handle the standard case where the features are correlation functions
    and the target is the user defined property (eg: bandgap, distortion etc)
    which is in the databse as key-value-pair.

    :param db_name: Name of the database being passed

    :param cf_names: List with the correlation function names to extract

    :param tab_name: Name of the table where the correlation functions are
        stored

    :param order: Order of the correlation function. Default 1.

    :prop: User given property
    """

    def __init__(
        self,
        prop: str,
        db_name: str,
        tab_name: str,
        cf_names: Optional[List[str]] = None,
        order: int = 1,
    ) -> None:
        super().__init__(db_name)
        self.tab_name = tab_name
        self.cf_names = cf_names
        self.order = order
        self.prop = prop

    def get_data(self, select_cond: List[tuple]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return X and y, where X is the design matrix containing correlation
        functions and y is the DFT energy per atom.

        :param select_cond: List with select conditions for the database
            (e.g. [('converged', '=', True)])
        """
        return self._get_data_from_getters(
            select_cond,
            CorrelationFunctionGetter(self.db_name, self.tab_name, self.cf_names, order=self.order),
            FinalStructPropertyGetter(self.db_name, self.prop),
        )


class CorrelationFunctionGetter(FeatureGetter):
    """
    CorrelationFunctionGetter is a class that extracts
    the correlation functions from an AtomsRow object

    :param db_name: Name of the database

    :param tab_name: Name of the external table where the correlation
        functions are stored

    :param cf_names: List with the names of the correlation functions.
        If None, all correlation functions in the database will be
        extracted

    :param order: Order of the correlation function. Default is 1.
    """

    def __init__(
        self, db_name: str, tab_name: str, cf_names: Optional[List[str]] = None, order: int = 1
    ) -> None:
        super().__init__(db_name)
        self.tab_name = tab_name
        self.cf_names = cf_names
        self.order = order

    @property
    def names(self):
        """
        Return a name of each column
        """
        return self.cf_names

    def _extract_cf_name(self, name: str) -> bool:
        """
        Return True if the name should be extracted
        """
        return self.cf_names is None or name in self.cf_names

    @staticmethod
    def _is_matrix_representable(id_cf_names: Dict[int, List[str]]) -> bool:
        """
        Check if the extracted correlation functions can be represented as a
        matrix.
        """
        reference = sorted(list(id_cf_names.values())[0])

        # Check the names of the correlation functions is equal for
        # all rows
        for v in id_cf_names.values():
            srt_v = sorted(v)
            if len(srt_v) != len(reference):
                return False
            if any(x != y for x, y in zip(reference, srt_v)):
                return False
        return True

    @staticmethod
    def _minimum_common_cf_set(id_cf_names: Dict[int, List[str]]) -> Set[str]:
        """
        Returns the minimum set of correlation functions that exists for all
        rows.
        """
        common_cf = set(list(id_cf_names.values())[0])
        for v in id_cf_names.values():
            common_cf = common_cf.intersection(v)
        return common_cf

    def get_property(self, ids: Sequence[int]) -> np.ndarray:
        """
        Extracts the design matrix associated with the database IDs. The first
        row in the matrix corresponds to the first item in ids, the second row
        corresponds to the second item in ids etc. If cf_names was None, all
        correlation functions in the database will be extracted. cf_names will
        be updated such that it reflects the names of the correlation functions
        that were extracted.

        :param ids: Database IDs of initial structures
        """
        sql = f"SELECT * FROM {self.tab_name}"
        id_cf_values = {}
        id_cf_names = {}

        id_set = set(ids)

        with self.connect() as db:
            self._check_version(db, *ids)
            cur = db.connection.cursor()
            cur.execute(sql)

            # Extract the correlation function name and value. The ID is also
            # extracted to check if it is in ids. The correlation functions
            # are placed in a dictionary because the IDs is not nessecarily
            # a monotone sequence of ints
            for name, value, db_id in cur.fetchall():
                if db_id in id_set and self._extract_cf_name(name):
                    cur_row = id_cf_values.get(db_id, [])
                    names = id_cf_names.get(db_id, [])

                    cur_row.append(value)
                    names.append(name)

                    id_cf_values[db_id] = cur_row
                    id_cf_names[db_id] = names

        if not self._is_matrix_representable(id_cf_names):
            min_set = self._minimum_common_cf_set(id_cf_names)
            msg = "The extracted correlation functions cannot be "
            msg += "represented as a matrix because there are a "
            msg += "number of columns for each row or at least one of "
            msg += "the names of the column differ.\n"
            msg += "Minimum common set of correlation functions.\n"
            msg += f"{min_set}\n"
            raise InconsistentDataError(msg)

        if self.cf_names is None:
            self.cf_names = set(v for item in id_cf_names.values() for v in item)
            self.cf_names = list(self.cf_names)
            self.cf_names = sort_cf_names(self.cf_names)

        # Make sure that all the rows are ordered in the same way
        cf_name_col = {n: i for i, n in enumerate(self.cf_names)}

        cf_matrix = np.zeros((len(id_cf_values), len(cf_name_col)))
        row = 0

        # Convert the correlation function dictionary to a numpy 2D array
        # such that is can be passed to a fitting algorithm
        for row, db_id in enumerate(ids):
            cf_values = id_cf_values[db_id]
            cf_names = id_cf_names[db_id]
            cols = [cf_name_col[name] for name in cf_names]
            cf_matrix[row, :] = np.array(cf_values)[cols]
            row += 1

        if self.order > 1:
            cf_matrix = self._include_higher_order(cf_matrix)
        return cf_matrix

    def _include_higher_order(self, cf_matrix: np.ndarray) -> np.ndarray:
        """
        Adds columns to the cf_matrix consisting of higher orders.

        :param cf_matrix: Matrix holding first order correlation functions
        """
        first_order_cf_indices = list(range(1, cf_matrix.shape[1]))
        nrows = cf_matrix.shape[0]
        for order in range(2, self.order + 1):
            for comb in cwr(first_order_cf_indices, r=order):
                X = np.ones((nrows, 1))
                for i in comb:
                    X[:, 0] *= cf_matrix[:, i]
                cf_matrix = np.hstack((cf_matrix, X))
                name = "*".join(self.cf_names[i] for i in comb)

                # Add the new to the list of cf names
                self.cf_names.append(name)
        return cf_matrix

    def _check_version(self, connection, *ids: Sequence[int]) -> None:
        if db_util.require_reconfigure_table(connection, self.tab_name, *ids):
            raise db_util.OutOfDateTable(
                "An out of date correlation function table was found. "
                'Please reconfigure your database, e.g. using "clease.reconfigure(settings)".'
            )


class FinalStructEnergyGetter(TargetGetter):
    """
    FinalStructEnergyGetter is a callable class that returns the final energy
    (typically after structure relaxation) corresponding to the passed
    AtomsRow object.

    :param db_name: Name of the database
    """

    @property
    def name(self):
        return "E_DFT (eV/atom)"

    def get_property(self, ids: Sequence[int]) -> np.ndarray:
        """
        Extract the final energy of the ids passed. In the returned array, the
        first energy corresponds to the first item in ids, the second energy
        corresponds to the second item in ids etc.

        :param ids: Database ids of initial structures
        """
        sql = "SELECT value, id FROM number_key_values "
        sql += "WHERE key='final_struct_id'"
        id_set = set(ids)
        # Map beetween the initial structure and the index in the id list
        init_struct_idx = {idx: i for i, idx in enumerate(ids)}

        with self.connect() as db:
            cur = db.connection.cursor()
            cur.execute(sql)
            final_struct_ids = []

            # Map the between final struct id and initial struct id
            init_struct_ids = {}
            for final_id, init_id in cur.fetchall():
                if init_id in id_set:
                    final_struct_ids.append(final_id)
                    init_struct_ids[final_id] = init_id

            # Extract the number of atoms of the initial structure
            num_atoms = extract_num_atoms(cur, id_set)

            final_struct_ids = set(final_struct_ids)
            sql = "SELECT id, energy FROM systems"
            cur.execute(sql)
            energies = np.zeros(len(final_struct_ids))
            # Populate the energies array
            for final_id, energy in cur.fetchall():
                if final_id in final_struct_ids:
                    init_id = init_struct_ids[final_id]
                    idx = init_struct_idx[init_id]
                    energies[idx] = energy / num_atoms[init_id]
        return energies


class FinalStructPropertyGetter(TargetGetter):
    """
    FinalStructPropertyGetter is a class that returns the user defined property
    value corresponding to the passed AtomsRow object.
    The user defined property should be located in the *final* atoms row.

    :param db_name: Name of the database
    """

    def __init__(self, db_name: str, prop: str) -> None:
        super().__init__(db_name)
        self.prop = prop

    @property
    def name(self):
        return f"Property: {self.prop}"

    def get_property(self, ids: Sequence[int]) -> np.ndarray:
        """
        Extract the property of the ids passed.

        :param ids: Database ids of initial structures
        """
        sql = "SELECT value, id FROM number_key_values "
        sql += "WHERE key='final_struct_id'"
        id_set = set(ids)
        # Map beetween the initial structure and the index in the id list
        init_struct_idx = {idx: i for i, idx in enumerate(ids)}

        with self.connect() as db:
            cur = db.connection.cursor()
            cur.execute(sql)
            final_struct_ids = []

            # Map the between final struct id and initial struct id
            init_struct_ids = {}
            for final_id, init_id in cur.fetchall():
                if init_id in id_set:
                    final_struct_ids.append(final_id)
                    init_struct_ids[final_id] = init_id

            final_struct_ids = set(final_struct_ids)
            sql = f"SELECT id, value FROM number_key_values WHERE key='{self.prop}'"
            cur.execute(sql)
            prop_values = np.zeros(len(final_struct_ids))

            # Populate the property array
            for final_id, prop_value in cur.fetchall():
                if final_id in final_struct_ids:
                    init_id = init_struct_ids[final_id]
                    idx = init_struct_idx[init_id]
                    prop_values[idx] = prop_value
        return prop_values


class FinalVolumeGetter(TargetGetter):
    """
    Class that extracts the final volume per atom of the relaxed structure

    :param db_name: Name of the database
    """

    @property
    def name(self):
        return "Volume (A^3)"

    def get_property(self, ids: Sequence[int]) -> np.ndarray:
        """
        Extracts the final volume of the ids passed. In the returned array,
        the first volume corresponds to the first item in ids, the second
        volume corresponds to the second item in ids etc.

        :param ids: Database IDs of initial structures
        """
        id_set = set(ids)
        query = "SELECT value, id FROM number_key_values "
        query += "WHERE key='final_struct_id'"

        init_ids = {}
        id_idx = {db_id: i for i, db_id in enumerate(ids)}
        with self.connect() as db:
            cur = db.connection.cursor()
            cur.execute(query)
            final_struct_ids = []

            # Extract the final struct id. Create a mapping between
            # final structure ids and initial structure ids for later
            # reference
            for value, db_id in cur.fetchall():
                final_id = int(value)
                if db_id in id_set:
                    final_struct_ids.append(final_id)
                    init_ids[final_id] = db_id

            num_atoms = extract_num_atoms(cur, id_set)

            final_struct_ids = set(final_struct_ids)
            query = "SELECT id, volume FROM systems"
            cur.execute(query)
            volumes = np.zeros(len(init_ids))

            # Extract the volume of the final structure and normalize it by
            # the number of atoms in the initial structure. The reason why
            # the number of atoms in the initial structure needs to be used,
            # is that vacancies are typically removed before the DFT run
            # starts. Consequently, they are missing in the final structure
            # and the number of atoms in that structure will be wrong
            for db_id, vol in cur.fetchall():
                if db_id in final_struct_ids:
                    init_id = init_ids[db_id]
                    idx = id_idx[init_id]
                    volumes[idx] = vol / num_atoms[init_id]
        return volumes


class CorrFuncVolumeDataManager(DataManager):
    """
    CorrFuncVolumeDataManager is a convenience class provided
    to handle the standard case where the features are correlation functions
    and the target is the volume of the relaxed cell

    :param db_name: Name of the database being passed

    :param tab_name: Name of the table where the correlation functions are
        stored

    :param cf_names: List with the correlation function names to extract.
        If None, all correlation functions in the database will be extracted.

    :param order: Order of the correlation functions. Default 1.
    """

    def __init__(
        self,
        db_name: str,
        tab_name: str,
        cf_names: Optional[List[str]] = None,
        order: int = 1,
    ) -> None:
        super().__init__(db_name)
        self.tab_name = tab_name
        self.cf_names = cf_names
        self.order = order

    def get_data(self, select_cond: List[tuple]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return X and y, where X is the design matrix containing correlation
        functions and y is the volume per atom.

        Parameters:

        select_cond: list
            List with select conditions for the database
            (e.g. [('converged', '=', True)])
        """
        # pylint: disable=arguments-differ
        return self._get_data_from_getters(
            select_cond,
            CorrelationFunctionGetter(self.db_name, self.tab_name, self.cf_names, order=self.order),
            FinalVolumeGetter(self.db_name),
        )


def extract_num_atoms(cur: sqlite3.Cursor, ids: Set[int]) -> Dict[int, int]:
    """
    Extract the number of atoms for all ids

    :param cur: SQL-cursor object
    :param ids: Set with IDs in the database.
    """
    sql = "SELECT id, natoms FROM systems"
    cur.execute(sql)
    num_atoms = {}
    for db_id, natoms in cur.fetchall():
        if db_id in ids:
            num_atoms[db_id] = natoms
    return num_atoms


class CorrelationFunctionGetterVolDepECI(DataManager):
    """
    Extracts correlation functions, multiplied with a power of the volume per
    atom. The feature names are named according to the correlation function
    names in the database, but a suffix of _Vd is appended. d is an integer
    inticading the power. Thus, if the name is for example c2_d0000_0_00_V2,
    it means that the column contains the correlation function c2_d0000_0_00,
    multiplied by V^2, where V is the volume per atom.

    :param db_name: Name of the database

    :param tab_name: Name of the table where correlation functions are stored

    :param cf_names: Name of the correlation functions that should be extracted

    :param order: Each ECI will be a polynomial in the volume of the passed
        order (default: 0)

    :param properties: List of properties that should be used in fitting. Can
        be energy, pressure, bulk_mod. (default: ['energy', 'pressure']).
        The pressure is always assumed to be zero (e.g. the energies passed
        are for relaxed structures.). All entries in the database are expected
        to have an energy. The remaining properties (e.g. bulk_mod) is not
        required for all structures. In class will pick up and the material
        property for the structures where it is present.

    :param cf_order: The energy is expanded up and (inluding) this order
        in the correlation function. Default is 1.
    """

    def __init__(
        self,
        db_name: str,
        tab_name: str,
        cf_names: List[str],
        order: Optional[int] = 0,
        properties: Tuple[str] = ("energy", "pressure"),
        cf_order: int = 1,
    ) -> None:
        super().__init__(db_name)
        self.tab_name = tab_name
        self.order = order
        self.cf_names = cf_names
        self.properties = properties
        self.cf_order = cf_order
        self._groups = []

    def build(self, ids: List[int]) -> np.ndarray:
        """
        Construct the design matrix and the target value required to fit a
        cluster expansion model to all material properties in self.properties.

        :param ids: List of ids to take into account
        """
        # pylint: disable=too-many-statements, too-many-locals
        cf_getter = CorrelationFunctionGetter(
            self.db_name, self.tab_name, self.cf_names, order=self.cf_order
        )
        cf = cf_getter.get_property(ids)

        volume_getter = FinalVolumeGetter(self.db_name)
        volumes = volume_getter.get_property(ids)

        energy_getter = FinalStructEnergyGetter(self.db_name)
        energies = energy_getter.get_property(ids)

        target_values = [energies]
        target_val_names = ["energy"]

        cf_vol_dep = np.zeros((cf.shape[0], (self.order + 1) * cf.shape[1]))
        counter = 0
        self._feat_names = []
        self._groups = list(range(cf_vol_dep.shape[0]))

        # Construct the part of the design matrix that corresponds to
        # energy fitting (e.g. if there are N energy calculations we construct
        # the firxt N rows of the design matrix)
        for col in range(cf.shape[1]):
            for power in range(self.order + 1):
                cf_vol_dep[:, counter] = cf[:, col] * volumes**power
                counter += 1
                self._feat_names.append(cf_getter.names[col] + f"_V{power}")

        # Add pressure data. P = dE/dV = 0 (we assume relaxed structures)
        if "pressure" in self.properties:
            pressure_cf = np.zeros_like(cf_vol_dep)
            counter = 0

            # Construct the part of the design matrix that corresponds to
            # pressure fitting (e.g. if there are N energy calculations,
            # we construct row N:2*N). Note that the pressure is assumed
            # to be zero
            for col in range(0, cf.shape[1]):
                for p in range(self.order + 1):
                    pressure_cf[:, counter] = p * cf[:, col] * volumes ** (p - 1)
                    counter += 1

            # Update the full design matrix and the target values
            cf_vol_dep = np.vstack((cf_vol_dep, pressure_cf))
            target_values.append(np.zeros(pressure_cf.shape[0]))
            target_val_names.append("pressure")

            # All data points have a pressure, the groups just repeats
            # themselves
            self._groups += self._groups

        id_row = {db_id: i for i, db_id in enumerate(ids)}

        # Add bulk modulus data. B = V*d^2E/dV^2
        if "bulk_mod" in self.properties:
            bulk_mod = self._extract_key(set(ids), "bulk_mod")
            bulk_mod_rows = [id_row[key] for key in bulk_mod]
            bulk_mod_cf = np.zeros((len(bulk_mod), cf_vol_dep.shape[1]))

            counter = 0
            vols = volumes[bulk_mod_rows]

            # Extract the part of the design matrix that corresponds to fitting
            # bulk moduli. If there are Nb calculations that has a bulk modulus
            # we construct row 2*N:2*N+Nb
            for col in range(0, cf.shape[1]):
                for power in range(self.order + 1):
                    bulk_mod_cf[:, counter] = vols * cf[bulk_mod_rows, col]
                    bulk_mod_cf[:, counter] *= power * (power - 1) * vols ** (power - 2)
                    counter += 1

            # Update the full design matrix as well as the target values
            cf_vol_dep = np.vstack((cf_vol_dep, bulk_mod_cf))
            target_values.append(bulk_mod.values())
            target_val_names.append("bulk_mod")
            self._groups += [self._groups[r] for r in bulk_mod_rows]

        # Add the pressure derivative of the bulk modulus (dB/dP)
        if "dBdP" in self.properties:
            dBdP = self._extract_key(set(ids), "dBdP")
            dBdP_rows = [id_row[key] for key in dBdP]
            dBdP_cf = np.zeros((len(dBdP), cf_vol_dep.shape[1]))
            dBdP = np.array(list(dBdP.values()))

            counter = 0
            vols = volumes[dBdP_rows]
            for col in range(cf.shape[1]):
                for power in range(self.order + 1):
                    # Double derivative of the energy
                    dE2 = power * (power - 1) * cf[dBdP_rows, col] * vols ** (power - 2)

                    # Third derivative of the energy
                    dE3 = (
                        power * (power - 1) * (power - 2) * cf[dBdP_rows, col] * vols ** (power - 3)
                    )
                    dBdP_cf[:, counter] = (1.0 + dBdP) * dE2 + vols * dE3
                    counter += 1

            # Update the full design matrix
            cf_vol_dep = np.vstack((cf_vol_dep, dBdP_cf))
            target_values.append(np.zeros(len(dBdP)))
            target_val_names.append("dBdP")
            self._groups += [self._groups[r] for r in dBdP_rows]

        # Assign the global design matrix to the attribute _X, the vector with
        # all target values to _y and construct the name of the target_value by
        # joining all subnames
        self._X = cf_vol_dep
        self._y = np.array([x for values in target_values for x in values])
        self._target_name = "-".join(target_val_names)

    def _extract_key(self, ids: Set[int], key: str) -> Dict[int, float]:
        """
        Extract a key from the database for the ids in the passed set that
        has a bulk modlus entry. The function returns a dictionary where the
        key is the ID in the database and the value is the extracted quantity

        :param ids: Set of all ids that should be considered

        :param key: Name of the key to be extracted
        """
        query = f"SELECT value, id FROM number_key_values WHERE key='{key}'"

        id_key = {}
        with connect(self.db_name) as db:
            cur = db.connection.cursor()
            cur.execute(query)
            for value, db_id in cur.fetchall():

                # Extract only the ones that are present in the set of ids
                if db_id in ids:
                    id_key[db_id] = value
        return id_key

    def get_data(self, select_cond: List[tuple]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return the design matrix and the target values for the entries
        corresponding to select_cond.

        :param select_cond: ASE select condition. The design matrix and the
            target vector will be extracted for rows matching the passed
            condition.
        """
        ids = get_ids(select_cond, self.db_name)
        self.build(ids)
        return self._X, self._y

    def groups(self) -> List[int]:
        """
        Return the group of each rows.
        """
        return self._groups


def make_corr_func_data_manager(
    prop: str, db_name: str, tab_name: str, cf_names: Sequence[str], **kwargs
) -> DataManager:
    """Helper function for creating a correlation function data manager"""
    if prop.lower() == "energy":
        return CorrFuncEnergyDataManager(db_name, tab_name, cf_names, **kwargs)
    # Assume it's a property data manager
    return CorrFuncPropertyDataManager(prop, db_name, tab_name, cf_names, **kwargs)



import logging
import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator
from .mc_observer import MCObserver

logger = logging.getLogger(__name__)


class LowestEnergyStructure(MCObserver):
    """Track the lowest energy state visited during an MC run.

    atoms: Atoms object
        Atoms object used in Monte Carlo

    verbose: bool
        If `True`, progress messages will be printed
    """

    name = "LowestEnergyStructure"

    def __init__(self, atoms, verbose=False):
        super().__init__()
        self.atoms = atoms
        self.lowest_energy = np.inf
        self.lowest_energy_cf = None
        self.emin_atoms = None
        self.verbose = verbose

    def reset(self):
        self.lowest_energy = np.inf
        self.lowest_energy_cf = None
        self.emin_atoms = None

    @property
    def calc(self):
        return self.atoms.calc

    @property
    def energy(self):
        return self.calc.results["energy"]

    def __call__(self, system_changes):
        """
        Check if the current state has lower energy and store the current
        state if it has a lower energy than the previous state.

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """

        if self.emin_atoms is None or self.energy < self.lowest_energy:
            self._update_emin()
            dE = self.energy - self.lowest_energy
            msg = "Found new low energy structure. "
            msg += f"New energy: {self.lowest_energy} eV. Change: {dE} eV"
            logger.info(msg)
            if self.verbose:
                print(msg)

    def _update_emin(self):
        self.lowest_energy_cf = self.calc.get_cf()
        self.lowest_energy = self.energy

        # Store emin atoms, and attach a cache calculator
        self.emin_atoms = self.atoms.copy()
        calc_cache = SinglePointCalculator(self.emin_atoms, energy=self.energy)
        self.emin_atoms.calc = calc_cache



from .mc_observer import MCObserver


class CorrelationFunctionObserver(MCObserver):
    """Track the history of the correlation function.

    Parameters:

    calc: `clease.calculators.Clease`
        Clease calculator

    names: list
        List with correlation functions to track.
        If None, all correlation functions are tracked.
    """

    name = "CorrelationFunctionObserver"

    def __init__(self, calc, names=None):
        super().__init__()
        current_cf = calc.get_cf()
        self.names = names
        if names is None:
            self.names = list(current_cf.keys())
        self.cf = {x: current_cf[x] for x in self.names}
        self.calc = calc
        self.counter = 1

    def __call__(self, system_changes):
        """Update the correlation functions.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        new_cf = self.calc.get_cf()
        for name in self.names:
            self.cf[name] += new_cf[name]
        self.counter += 1

    def reset(self):
        self.cf = self.calc.get_cf()
        self.counter = 1

    def get_averages(self):
        return {"cf_" + k: v / self.counter for k, v in self.cf.items()}



import numpy as np

__all__ = ("name_clusters", "get_first_match", "diameter", "size")


def name_clusters(fingerprints):
    """Name clusters based on its size and diameter."""
    distance_list = [2 * np.sqrt(fp[0]) for fp in fingerprints]

    distance_list = np.round(distance_list, decimals=6)
    distance_list = np.unique(distance_list)
    distance_list.sort()
    prefixes = []
    prefix_dict = {}
    for i, fp in enumerate(fingerprints):
        match = get_first_match(fp, fingerprints[:i])
        s = size(fp)
        if match != -1:
            prefix = prefixes[match]
        else:
            dia = diameter(fp)
            index = np.argmin(np.abs(distance_list - dia))
            assert abs(dia - distance_list[index]) < 2e-6

            if index < 10:
                prefix = f"c{s}_d000{index}"
            elif index < 100:
                prefix = f"c{s}_d00{index}"
            elif index < 1000:
                prefix = f"c{s}_d0{index}"
            else:
                prefix = f"c{s}_d{index}"
        prefixes.append(prefix)

        fp_list = prefix_dict.get(prefix, [])
        if fp not in fp_list:
            fp_list.append(fp)
        prefix_dict[prefix] = fp_list

    # Sort prefix dict
    for _, v in prefix_dict.items():
        v.sort()

    names = []
    for prefix, fp in zip(prefixes, fingerprints):
        suffix = prefix_dict[prefix].index(fp)
        names.append(prefix + f"_{suffix}")
    return names


def get_first_match(fingerprint, fingerprints):
    """
    Find index of the first instance where the passed fingerprint is within the
    list of the fingerprints.
    """
    for i, fp in enumerate(fingerprints):
        if fp == fingerprint:
            return i
    return -1


def diameter(fingerprint):
    """Get diameter of a cluster using its fingerprint."""
    return 2 * np.sqrt(fingerprint[0])


def size(fingerprint):
    """Get size of the cluster using its fingerprint."""
    num = np.sqrt(2 * len(fingerprint) + 0.25)
    s = num - 0.5
    assert abs(int(s) - s) < 1e-9
    return int(s)
